import numpy as np, numba as nb
from multiprocessing import shared_memory

class getDistance(object) :
    def __init__(self, data, func_name, pool, start=0):
        func = eval(func_name)
        self.mat_buf = shared_memory.SharedMemory(create=True, size=data.nbytes)
        mat = np.ndarray(data.shape, dtype=data.dtype, buffer=self.mat_buf.buf)
        mat[:] = data[:]
        self.dist_buf = shared_memory.SharedMemory(create=True, size=int((mat.shape[0] - start) * mat.shape[0] * 4))
        self.dist = np.ndarray([mat.shape[0] - start, mat.shape[0]], dtype=np.int32, buffer=self.dist_buf.buf)
        self.dist[:] = 0
        parallel_dist(self.mat_buf, func, self.dist_buf, mat.shape, pool, start)
        self.mat_buf.close()
        self.mat_buf.unlink()
    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dist_buf.close()
        self.dist_buf.unlink()
        return


def parallel_dist(mat_buf, func, dist_buf, mat_shape, pool, start=0) :
    n_pool = len(pool._pool)
    tot_cmp = (mat_shape[0] * mat_shape[0] - start * start)/n_pool
    s, indices = start, []
    for _ in np.arange(n_pool) :
        e = np.sqrt(s*s+tot_cmp)
        indices.append([s, e])
        s = e
    indices = (np.array(indices)+0.5).astype(int)
    for _ in pool.imap_unordered(dist_wrapper, [[func, mat_buf.name, dist_buf.name, mat_shape, s, e, start] for s, e in indices ]) :
        pass

def dist_wrapper(data) :
    func, mat_buf_id, dist_buf_id, mat_shape, s, e, start = data
    mat_buf = shared_memory.SharedMemory(name=mat_buf_id, create=False)
    dist_buf = shared_memory.SharedMemory(name=dist_buf_id, create=False)
    mat = np.ndarray(mat_shape, dtype=int, buffer=mat_buf.buf)
    dist = np.ndarray([mat_shape[0]-start, mat_shape[0]], dtype=np.int32, buffer=dist_buf.buf)
    func(mat[:, 1:], s, e, dist, start)
    mat_buf.close()
    dist_buf.close()

@nb.jit(nopython=True)
def syn_dist(mat, s, e, dist, start=0):
    n_loci = mat.shape[1]
    for i in range(s, e+1) :
        ql = np.sum(mat[i] > 0)
        for j in range(i) :
            rl, ad, al = 0., 0., 0.
            for k in range(n_loci) :
                if mat[j, k] > 0 :
                    rl += 1
                    if mat[i, k] > 0 :
                        al += 1
                        if mat[i, k] != mat[j, k] :
                            ad += 1
            ll = max(ql, rl) - 0.03 * n_loci
            if ll > al :
                ad += ll - al
                al = ll
            if al > 0 :
                dist[i-start, j] = int(ad/al * n_loci + 0.5)
            else :
                dist[i-start, j] = int(n_loci)

@nb.jit(nopython=True)
def asyn_dist(mat, s, e, dist, start=0):
    n_loci = mat.shape[1]
    for i in range(s, e+1) :
        ql = np.sum(mat[i] > 0)
        for j in range(i) :
            rl, ad, al = 0., 0., 0.
            for k in range(n_loci) :
                if mat[j, k] > 0 :
                    if mat[i, k] > 0 :
                        al += 1
                        if mat[i, k] != mat[j, k] :
                            ad += 1
            ll = ql - 0.03 * n_loci
            if ll > al :
                ad += ll - al
                al = ll
            if al > 0 :
                dist[i-start, j] = int(ad/al * n_loci + 0.5)
            else :
                dist[i-start, j] = int(n_loci)

@nb.jit(nopython=True)
def p_dist(mat, s, e, dist, start=0):
    n_loci = mat.shape[1]
    for i in range(s, e+1) :
        for j in range(i) :
            ad, al = 0., 0.
            for k in range(n_loci) :
                if mat[j, k] > 0 :
                    if mat[i, k] > 0 :
                        al += 1
                        if mat[i, k] != mat[j, k] :
                            ad += 1
            dist[i-start, j] = int( -np.log(1.-(ad+0.5)/(al+1.0)) * n_loci * 100. + 0.5)



#!/usr/bin/env python

# HierCC.py
# Hierarchical Clustering Complex of MLST allelic profiles
#
# Author: Zhemin Zhou
# Lisence: GPLv3
#
# New assignment: hierCC.py -p <allelic_profile> -o <output_prefix>
# Incremental assignment: hierCC.py -p <allelic_profile> -o <output_prefix> -i <old_cluster_npy>
# Input format:
# ST_id gene1 gene2
# 1 1 1
# 2 1 2
# ...

import sys, gzip, logging, click
import pandas as pd, numpy as np
from multiprocessing import Pool, set_start_method
from scipy.spatial import distance as ssd
from scipy.cluster.hierarchy import linkage
try :
    from getDistance import getDistance
except :
    from .getDistance import getDistance

logging.basicConfig(format='%(asctime)s | %(message)s',stream=sys.stdout, level=logging.INFO)

def prepare_mat(profile_file) :
    mat = pd.read_csv(profile_file, sep='\t', header=None, dtype=str).values
    allele_columns = np.array([i == 0 or (not h.startswith('#')) for i, h in enumerate(mat[0])])
    mat = mat[1:, allele_columns].astype(int)
    mat = mat[mat.T[0]>0]
    return mat

@click.command()
@click.option('-p', '--profile', help='[INPUT; REQUIRED] name of the profile file. Can be GZIPed.',
                        required=True)
@click.option('-o', '--output',
                        help='[OUTPUT; REQUIRED] Prefix for the output files. These include a NUMPY and TEXT verions of the same clustering result',
                        required=True)
@click.option('-a', '--append', help='[INPUT; optional] The NUMPY version of an existing HierCC result',
                        default='')
@click.option('-n', '--n_proc', help='[DEFAULT: 4] Number of processors.', default=4, type=int)
def hierCC(profile, output, append, n_proc):
    '''HierCC takes allelic profile (as in https://pubmlst.org/data/) and
    work out hierarchical clusters of all the profiles based on a minimum-spanning tree.'''
    pool = Pool(n_proc)

    profile_file, cluster_file, old_cluster = profile, output + '.npz', append

    mat = prepare_mat(profile_file)
    n_loci = mat.shape[1] - 1

    logging.info(
        'Loaded in allelic profiles with dimension: {0} and {1}. The first column is assumed to be type id.'.format(
            *mat.shape))
    logging.info('Start hierCC assignments')

    # prepare existing clusters
    if not append:
        absence = np.sum(mat <= 0, 1)
        mat[:] = mat[np.argsort(absence, kind='mergesort')]
        typed = {}
    else :
        od = np.load(old_cluster, allow_pickle=True)
        cls = od['hierCC']
        typed = {c: id for id, c in enumerate(cls.T[0]) if c > 0}
    if len(typed) > 0:
        logging.info('Loaded in {0} old hierCC assignments.'.format(len(typed)))
        mat_idx = np.array([t in typed for t in mat.T[0]])
        mat[:] = np.vstack([mat[mat_idx], mat[(mat_idx) == False]])
        start = np.sum(mat_idx)
    else :
        start = 0

    res = np.repeat(mat.T[0], int(mat.shape[1]) + 1).reshape(mat.shape[0], -1)
    res[res < 0] = np.max(mat.T[0]) + 100
    res.T[0] = mat.T[0]
    logging.info('Calculate distance matrix')
    # prepare existing tree
    if append :
        for r in res :
            if r[0] in typed :
                r[:] = cls[typed[r[0]]]
    else :
        with getDistance(mat, 'syn_dist', pool, start) as dist :
            dist.dist += dist.dist.T
            logging.info('Start Single linkage clustering')
            slc = linkage(ssd.squareform(dist.dist), method='single')

        index = { s:i for i, s in enumerate(mat.T[0]) }
        descendents = [ [m] for m in mat.T[0] ] + [None for _ in np.arange(mat.shape[0]-1)]
        for idx, c in enumerate(slc.astype(int)) :
            n_id = idx + mat.shape[0]
            d = sorted([int(c[0]), int(c[1])], key=lambda x:descendents[x][0])
            min_id = min(descendents[d[0]])
            descendents[n_id] = descendents[d[0]] + descendents[d[1]]
            for tgt in descendents[d[1]] :
                res[index[tgt], c[2]+1:] = res[index[min_id], c[2]+1:]
    logging.info('Attach genomes onto the tree.')
    with getDistance(mat, 'asyn_dist', pool, start) as dist :
        for id, (r, d) in enumerate(zip(res[start:], dist.dist)):
            if id + start > 0 :
                i = np.argmin(d[:id+start])
                min_d = d[i]
                if r[min_d + 1] > res[i, min_d + 1]:
                    r[min_d + 1:] = res[i, min_d + 1:]
    #pool.close()
    res.T[0] = mat.T[0]
    np.savez_compressed(cluster_file, hierCC=res)

    with gzip.open(output + '.hierCC.gz', 'wt') as fout:
        fout.write('#ST_id\t{0}\n'.format('\t'.join(['HC' + str(id) for id in np.arange(n_loci+1)])))
        for r in res[np.argsort(res.T[0])]:
            fout.write('\t'.join([str(rr) for rr in r]) + '\n')

    logging.info('NUMPY clustering result (for incremental hierCC): {0}.npz'.format(output))
    logging.info('TEXT  clustering result (for visual inspection): {0}.hierCC.gz'.format(output))
    pool.close()

if __name__ == '__main__':
    set_start_method('spawn')
    hierCC(sys.argv[1:])




#!/usr/bin/env python

import sys, pandas as pd, numpy as np, logging, sysv_ipc
import click
from sklearn.metrics import silhouette_score, normalized_mutual_info_score
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from multiprocessing import Pool
try :
    from getDistance import getDistance
except :
    from .getDistance import getDistance

logging.basicConfig(format='%(asctime)s | %(message)s',stream=sys.stdout, level=logging.INFO)


def get_similarity2(data) :
    method, cc1, cc2 = data
    if np.unique(cc1).size == 1 and  np.unique(cc1).size == 1 :
        return 1.
    return method(cc1, cc2)

def get_similarity(method, cluster, stepwise, pool) :
    logging.info('Calculating NMIs...')
    similarity = np.ones([cluster.shape[1], cluster.shape[1]], dtype=np.float64)
    for i1, cc1 in enumerate(cluster.T) :
        if i1 % 10 == 0 :
            logging.info('    NMIs between level {0} and greater levels'.format(i1 * stepwise))
        similarity[i1, i1+1:] = pool.map(get_similarity2, [ [method, cc1, cc2] for cc2 in cluster.T[i1+1:] ])
        similarity[i1+1:, i1] = similarity[i1, i1+1:]
    similarity[similarity>0.999] = 0.999
    similarity[similarity<0.0] = 0.0
    return similarity

def get_silhouette(profile, cluster, stepwise, pool) :
    logging.info('Calculating pairwise distance ...')
    with getDistance(profile, 'p_dist', pool) as dist :
        dist.dist += dist.dist.T
        logging.info('Calculating Silhouette score ...')
        silhouette = np.array(pool.map(get_silhouette2, [ [dist.dist_buf.key, dist.dist.shape, tag] for tag in cluster.T ]))
    return silhouette

def get_silhouette2(data) :
    dist_key, dist_shape, tag = data
    s = np.unique(tag)
    if 2 <= s.size < tag.shape[0] :
        dist_buf = sysv_ipc.SharedMemory(dist_key)
        dist = np.ndarray(dist_shape, dtype=np.int32, buffer=memoryview(dist_buf))
        ss = silhouette_score(dist.astype(float), tag, metric = 'precomputed')
        dist_buf.detach()
        return ss
    else :
        return 0.

def prepare_mat(profile_file) :
    mat = pd.read_csv(profile_file, sep='\t', header=None, dtype=str).values
    allele_columns = np.array([i == 0 or (not h.startswith('#')) for i, h in enumerate(mat[0])])
    mat = mat[1:, allele_columns].astype(int)
    mat = mat[mat.T[0]>0]
    return mat


@click.command()
@click.option('-p', '--profile', help='[INPUT; REQUIRED] name of the profile file. Can be GZIPed.', required=True)
@click.option('-c', '--cluster', help='[INPUT; REQUIRED] name of the HierCC file. Can be GZIPed.', required=True)
@click.option('-o', '--output', help='[OUTPUT; REQUIRED] Prefix for the output files.', required=True)
@click.option('-s', '--stepwise', help='[DEFAULT: 10] Evaluate every <stepwise> levels.', default=10, type=int)
@click.option('-n', '--n_proc', help='[DEFAULT: 4] Number of processors.', default=4, type=int)
def evalHCC(profile, cluster, output, stepwise, n_proc) :
    '''evalHCC evaluates HierCC results using varied statistic summaries.'''
    pool = Pool(n_proc)

    profile = prepare_mat(profile)
    cluster = prepare_mat(cluster)

    idx = { p:i for i, p in enumerate(profile.T[0])}
    cluster_idx = sorted([ [idx.get(c, -1), i] for i, c in enumerate(cluster.T[0]) if c in idx ])
    cluster = cluster[np.array(cluster_idx).T[1]]

    idx = { p:i for i, p in enumerate(cluster.T[0])}
    cluster_idx = sorted([ [idx.get(c, -1), i] for i, c in enumerate(profile.T[0]) if c in idx ])
    profile = profile[np.array(cluster_idx).T[1]]
    cluster = cluster[:, 1::stepwise]

    silhouette = get_silhouette(profile, cluster, stepwise, pool)
    similarity = get_similarity(normalized_mutual_info_score, cluster, stepwise, pool)

    with open(output+'.tsv', 'w') as fout:
        levels = ['HC{0}'.format(lvl*stepwise) for lvl in np.arange(silhouette.shape[0])]
        for lvl, ss in zip(levels, silhouette) :
            fout.write('#Silhouette\t{0}\t{1}\n'.format(lvl, ss))

        fout.write('\n#NMI\t{0}\n'.format('\t'.join(levels)))
        for lvl, nmis in zip(levels, similarity):
            fout.write('{0}\t{1}\n'.format(lvl, '\t'.join([ '{0:.3f}'.format(nmi) for nmi in nmis ])))
    fig, axs = plt.subplots(2, 2, \
                            figsize=(8, 12), \
                            gridspec_kw={'width_ratios':(12, 1),
                                         'height_ratios': (65, 35)})

    heatplot = axs[0, 0].imshow( (10*(np.log10(1-similarity))), \
                                norm=colors.TwoSlopeNorm(vmin=-30., vcenter=-10., vmax=0), \
                                cmap = 'RdBu',\
                                extent=[0, silhouette.shape[0]*stepwise, \
                                        silhouette.shape[0]*stepwise, 0])
    cb = fig.colorbar(heatplot, cax=axs[0, 1])
    axs[1, 0].plot(np.arange(silhouette.shape[0])*stepwise, silhouette,)
    axs[1, 0].set_xlim([0, silhouette.shape[0]*stepwise])
    axs[1, 1].remove()
    axs[0, 0].set_ylabel('HCs (allelic distances)')
    axs[0, 0].set_xlabel('HCs (allelic distances)')
    axs[1, 0].set_ylabel('Silhouette scores')
    axs[1, 0].set_xlabel('HCs (allelic distances)')
    cb.set_label('Normalized Mutual Information')
    cb.set_ticks([-30, -23.01, -20, -13.01, -10, -3.01, 0])
    cb.ax.set_yticklabels(['>=.999', '.995', '.99', '.95', '.9', '.5', '.0'])
    plt.savefig(output+'.pdf')
    logging.info('Tab delimited evaluation is save in {0}.tsv'.format(output))
    logging.info('Graphic visualisation is save in {0}.pdf'.format(output))

if __name__ == '__main__' :
    evalHCC()



import os, sys
from setuptools import setup, find_packages
#from PEPPA import __VERSION__
__VERSION__ = '1.1'

with open('README.rst', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='HierCC',
    version= __VERSION__,
    #scripts=['PEPPA.py'] ,
    author="Zhemin Zhou",
    author_email="zhemin.zhou@warwick.ac.uk",
    description="Hierarchical Clustering of cgMLST profiles",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zheminzhou/HierCC",
    packages = ['HierCC'],
    package_dir = {'HierCC':'.'},
    keywords=['bioinformatics', 'microbial', 'genomics', 'MLST', 'hierarhical clustering'],
    install_requires=['matplotlib>=3.2.1', 'scikit-learn>=0.23.1', 'numba>=0.38.0', 'numpy>=1.18.1', 'pandas>=0.24.2', 'scipy>=1.3.2'],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'HierCC = HierCC.HierCC:hierCC',
            'HCCeval = HierCC.HCCeval:evalHCC',
    ]},
    package_data={'HierCC': ['LICENSE', 'README.*']},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        "Programming Language :: Python :: 3.8",
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        "Operating System :: OS Independent",
    ],
 )

